from pathlib import Path
from typing import Optional, Dict

from .db import entity_generator, repository_generator, adapter_generator
from .config_context import get_config
from crudhex.domain.models import Entity


def create_entity_class(entity: Entity, folder: Optional[Path] = None) -> Path:
    if not folder:
        folder = Path(get_config().get_db_models_path())

    return entity_generator.create_entity_class(entity, folder)


def create_repository_class(entity: Entity, folder: Optional[Path] = None) -> Path:
    if not folder:
        folder = Path(get_config().get_db_repositories_path())

    return repository_generator.create_repository_class(entity, folder)


def create_adapter_class(entity: Entity, entities_map: Dict[str, Entity], folder: Optional[Path] = None) -> Path:
    if not folder:
        folder = Path(get_config().get_db_adapters_path())

    return adapter_generator.create_adapter_class(entity, entities_map, folder)



from pathlib import Path
from typing import List, Optional

from crudhex.domain.models import Entity
from crudhex.domain.utils.class_type_utils import get_import, get_field_imports
from ..config_context import get_config
from ..domain import model_generator, command_generator, use_case_port_generator

from crudhex.domain.ports import rest_code_writer

_CONTROLLER_SUFFIX = 'Controller'


def create_controller_class(entity: Entity, folder: Optional[Path] = None) -> Path:
    if not folder.is_dir(): raise RuntimeError('Output path must be a folder ({})'.format(folder.resolve()))

    class_type = get_controller_type_name(entity)
    controller_file = folder / f'{class_type}.java'

    model_type = model_generator.get_model_type_name(entity)
    id_type = entity.get_id_field().type.class_type
    use_case_type = use_case_port_generator.get_port_type_name(entity)
    command_type = command_generator.get_command_type_name(entity)

    rest_code_writer.create_controller(
        controller_file, class_type, get_package(), _get_imports(entity),
        entity.name, model_type, id_type,
        use_case_type, command_type, command_type
    )

    return controller_file


def get_controller_type_name(entity: Entity) -> str:
    return f'{entity.name}{_CONTROLLER_SUFFIX}'


def get_package() -> str:
    return get_config().rest_controllers_pkg


def _get_imports(entity: Entity) -> List[str]:
    imports = [
        get_import(model_generator.get_package(), model_generator.get_model_type_name(entity)),
        get_import(command_generator.get_package(), command_generator.get_command_type_name(entity)),
        get_import(use_case_port_generator.get_package(), use_case_port_generator.get_port_type_name(entity))
    ]
    imports += get_field_imports(entity.get_id_field())

    return imports




from pathlib import Path
from typing import List

from crudhex.adapters.infrastructure.template_writer.config import template_config
from crudhex.adapters.infrastructure.template_writer.config.template_config import get_rest_file_path
from crudhex.adapters.infrastructure.template_writer.services.template_env import get_template_environment


def create_controller(dest: Path, class_type: str, package: str,
                      imports: List[str], entity_name: str, model_type: str, id_type: str,
                      use_case_type: str, create_command_type: str, update_command_type: str):

    template_env = get_template_environment()

    repository_template = template_env.get_template(get_rest_file_path(template_config.CONTROLLER_TEMPLATE))
    repository_code = repository_template.render({
        'package': package,
        'imports': '\n'.join(imports),
        'class_type': class_type,
        'name': entity_name,
        'use_case_type': use_case_type,
        'model_type': model_type,
        'id_type': id_type,
        'create_command_type': create_command_type,
        'update_command_type': update_command_type
    })

    with open(dest.resolve(), 'w+', encoding='utf-8') as f:
        f.write(repository_code)



from pathlib import Path
from typing import List, Optional

from crudhex.domain.models import Entity
from crudhex.domain.utils import package_utils
from ..config_context import get_config
from ..type_resolver import get_type_resolver
from . import entity_generator as entity_generator

from crudhex.domain.ports import db_code_writer

_DB_REPO_SUFFIX = 'Repository'


def create_repository_class(entity: Entity, folder: Optional[Path] = None) -> Path:
    if not folder.is_dir(): raise RuntimeError('Output path must be a folder ({})'.format(folder.resolve()))

    class_type = get_repository_type_name(entity)
    repo_file = folder / f'{class_type}.java'

    db_code_writer.create_entity_repository(
        repo_file, class_type, get_package(), _get_imports(entity),
        entity_generator.get_entity_type_name(entity), entity.get_id_field().type.class_type
    )

    return repo_file


def get_repository_type_name(entity: Entity) -> str:
    return f'{entity.name}{_DB_REPO_SUFFIX}'


def get_package() -> str:
    return get_config().db_repositories_pkg


def _get_imports(entity: Entity) -> List[str]:
    type_resolver = get_type_resolver()

    class_types = [
        package_utils.full_class_name(entity_generator.get_package(), entity_generator.get_entity_type_name(entity))
    ]
    class_types.extend(type_resolver.get_field_types_full_class(entity.get_id_field()))

    return [package_utils.generate_import(ct) for ct in class_types]




from pathlib import Path
from typing import Optional

from crudhex.domain.models import Entity
from crudhex.domain.services.config_context import get_config
from crudhex.domain.services.rest import controller_generator


def create_model_class(entity: Entity, folder: Optional[Path] = None) -> Path:
    if not folder:
        folder = Path(get_config().get_rest_controllers_path())

    return controller_generator.create_controller_class(entity, folder)



from pathlib import Path
from typing import List, Dict

from crudhex.domain.models import Entity, Field
from crudhex.domain.utils.class_type_utils import get_field_imports, get_field_types
from ..config_context import get_config

from crudhex.domain.ports import domain_code_writer

_COMMAND_SUFFIX = 'UpsertCommand'


def create_command_class(entity: Entity, entities_map: Dict[str, Entity], folder: Path) -> Path:
    if not folder.is_dir(): raise RuntimeError('Output path must be a folder ({})'.format(folder.resolve()))

    class_type = get_command_type_name(entity)
    model_file = folder / f'{class_type}.java'

    domain_code_writer.create_command(model_file, get_command_type_name(entity), get_package(),
                                      _get_command_imports(entity), _get_command_fields_data(entity, entities_map))

    return model_file


def get_package() -> str:
    return get_config().domain_commands_pkg


def get_command_type_name(entity: Entity) -> str:
    return f'{entity.name}{_COMMAND_SUFFIX}'


def get_command_used_fields(entity: Entity) -> List[Field]:
    return [f for f in entity.fields if _is_used_field(f)]


def _get_command_imports(entity: Entity) -> List[str]:
    imports = []
    for field in entity.fields:
        if not _is_used_field(field): continue
        imports += get_field_imports(field)

    return imports


def _get_command_fields_data(entity: Entity, entities_map: Dict[str, Entity]) -> List[Dict[str, str]]:
    fields_data = []
    for field in entity.fields:
        # Exclude non owning relation side getters to avoid loops in serialization
        if not _is_used_field(field): continue
        fields_data.append(_get_field_data(field, entities_map))

    return fields_data


def _get_field_data(field: Field, entities_map: Dict[str, Entity]) -> Dict[str, str]:
    field_data = {
        'name': field.name
    }
    if field.has_relation():
        related_entity = entities_map.get(field.type.class_type)
        field = _get_relation_id_field_data(related_entity, field)

    field_data.update(get_field_types(field)._asdict())

    return field_data


def _get_relation_id_field_data(entity: Entity, relation_field: Field) -> Field:
    id_field = entity.get_id_field()
    if not id_field: raise RuntimeError('No id field found for entity {}'.format(entity.name))

    id_relation_field = Field(relation_field.name, id_field.type.class_type, False)
    id_relation_field.type.collection_type = relation_field.type.collection_type

    return id_relation_field


def _is_used_field(field: Field) -> bool:
    inverse_relation = field.has_relation() and field.relation.mapped_by
    return not inverse_relation and not field.is_id()



from pathlib import Path
from typing import Optional, Set, Union, List

from crudhex.domain.config.SpecConfig import SpecConfig
from crudhex.domain.models import IdMeta, Relation, GenerationType, RelationType
from crudhex.domain.models import Entity, Field
from crudhex.domain.services import TypeResolver
from crudhex.domain.services.type_resolver import get_type_resolver

from crudhex.domain.ports import spec_loader


def parse_spec_file(spec_path: Path) -> List[Entity]:
    spec_data = spec_loader.load_spec_config(spec_path)
    type_resolver = get_type_resolver()

    return parse_spec_data(spec_data, type_resolver)


def parse_spec_data(spec_data: dict, type_resolver: Optional[TypeResolver] = None) -> List[Entity]:
    entities = []
    # noinspection PyTypeChecker
    generated_types = spec_data.keys()  # type: set
    for name, spec in spec_data.items():
        entities.append(parse_entity_data(name, spec, generated_types, type_resolver))

    return entities


def parse_entity_data(name: str, data: dict, generated_types: Set[str],
                      type_resolver: Optional[TypeResolver]) -> Entity:
    entity = Entity(name)
    entity.meta = parse_entity_meta(data)

    fields_data = data.copy()
    if SpecConfig.META in fields_data: fields_data.pop(SpecConfig.META)

    for name, field_data in fields_data.items():
        entity.fields.append(parse_field(name, field_data, generated_types, type_resolver))

    return entity


def parse_entity_meta(data: dict) -> Optional[Entity.Meta]:
    meta = data.get(SpecConfig.META)
    if not meta or SpecConfig.TABLE not in meta: return None

    return Entity.Meta(meta[SpecConfig.TABLE])


def parse_field(name: str, data: Union[str, dict], generated_types: Set[str],
                type_resolver: Optional[TypeResolver]) -> Field:
    if type(data) is str:
        class_type = data if not type_resolver else type_resolver.resolve_type(data)
        return Field(name, class_type, class_type in generated_types)

    if SpecConfig.TYPE not in data: raise RuntimeError(f'Missing required key {SpecConfig.TYPE}')

    class_type = (data[SpecConfig.TYPE]
                  if not type_resolver
                  else type_resolver.resolve_type(data[SpecConfig.TYPE]))

    field = Field(name, class_type, class_type in generated_types)
    field.column = data.get(SpecConfig.COLUMN)
    field.id_meta = parse_field_id_meta(data)
    field.relation = parse_field_relation(data)

    # TODO: Configurable in dsl file
    if field.has_relation() and field.relation.type.has_multiple():
        field.type.collection_type = TypeResolver.DEFAULT_COLLECTION

    return field


def parse_field_id_meta(data: dict) -> Optional[IdMeta]:
    id_data = data.get(SpecConfig.ID)
    if not id_data: return None

    if type(id_data) is str:
        return IdMeta(GenerationType(id_data))

    id_meta = IdMeta(GenerationType(id_data[SpecConfig.TYPE]))
    id_meta.sequence = id_data.get(SpecConfig.SEQUENCE)

    return id_meta


def parse_field_relation(data: dict) -> Optional[Relation]:
    relation_data = data.get(SpecConfig.RELATION)
    if not relation_data: return None

    relation = Relation(RelationType(relation_data[SpecConfig.TYPE]))
    relation.join_table = relation_data.get(SpecConfig.JOIN_TABLE)
    relation.join_column = relation_data.get(SpecConfig.JOIN_COLUMN)
    relation.inverse_join_column = relation_data.get(SpecConfig.INVERSE_JOIN_COLUMN)
    relation.mapped_by = relation_data.get(SpecConfig.MAPPED_BY)

    return relation



from pathlib import Path
from typing import Optional

from crudhex.domain.models import ProjectConfig
from crudhex.domain.ports import spec_loader

DEFAULT_CONFIG = './.crudhex-conf.yaml'
_CONFIG: Optional[ProjectConfig] = None


def load_config(path: Optional[Path] = None):
    if not path: path = Path(DEFAULT_CONFIG)

    config = spec_loader.load_project_config(path)
    set_config(config)

    return config


def set_config(config: ProjectConfig):
    global _CONFIG
    _CONFIG = config


def get_config(default_config: Optional[ProjectConfig] = None) -> ProjectConfig:
    """
    Returns config in context. If not available it will try to use provided `default_config`, otherwise will raise an exeption
    :param default_config:
    :raise RuntimeError: If no config nor default available
    :return:
    """
    global _CONFIG
    if not _CONFIG:
        if not default_config: raise RuntimeError("Config not available in context")
        else: return default_config

    return _CONFIG


def has_config() -> bool:
    global _CONFIG
    return _CONFIG is not None



from pathlib import Path
from typing import List, Dict

from crudhex.domain.models import Entity, Field
from crudhex.domain.utils.class_type_utils import get_field_imports, get_field_types
from ..config_context import get_config

from crudhex.domain.ports import domain_code_writer


def create_model_class(entity: Entity, folder: Path) -> Path:
    if not folder.is_dir(): raise RuntimeError('Output path must be a folder ({})'.format(folder.resolve()))

    class_type = get_model_type_name(entity)
    model_file = folder / f'{class_type}.java'

    domain_code_writer.create_model(model_file, get_model_type_name(entity), get_package(),
                                    _get_model_imports(entity), _get_model_fields_data(entity))

    return model_file


def get_package() -> str:
    return get_config().domain_models_pkg


def get_model_type_name(entity: Entity) -> str:
    return entity.name


def _get_model_imports(entity: Entity) -> List[str]:
    imports = []
    for field in entity.fields:
        # Exclude non owning relation side getters to avoid loops in serialization
        if not _is_used_field(field): continue
        imports += get_field_imports(field)

    return imports


def _get_model_fields_data(entity: Entity) -> List[Dict[str, str]]:
    fields_data = []
    for field in entity.fields:
        # Exclude non owning relation side getters to avoid loops in serialization
        if not _is_used_field(field): continue
        fields_data.append(_get_field_data(field))

    return fields_data


def _get_field_data(field: Field) -> Dict[str, str]:
    field_data = {
        'name': field.name
    }
    field_data.update(get_field_types(field)._asdict())

    return field_data


def _is_used_field(field: Field) -> bool:
    return not field.has_relation() or not field.relation.mapped_by



from pathlib import Path
from typing import Tuple, Optional


IMPORT_TEMPLATE = 'import {};'


def pkg_to_path(pkg: str, base_path: str = '.') -> Path:
    pkg_path = pkg.replace('.', '/')

    return Path(base_path) / pkg_path


def is_full_class_name(class_name: str) -> bool:
    return len(class_name.split('.')) > 1


def get_class_name(class_name: str) -> str:
    return class_name.split('.')[-1]


def parse_class_name(class_name: str) -> Tuple[str, str]:
    class_pkgs = class_name.split('.')

    pkg = None if len(class_pkgs) == 1 else '.'.join(class_pkgs[0:-1])
    cls_type = class_pkgs[-1]

    return pkg, cls_type


def full_class_name(pkg: Optional[str], class_type: str):
    if not pkg: return class_type

    return '.'.join((pkg, class_type))


def generate_import(full_class_name: str) -> str:
    return IMPORT_TEMPLATE.format(full_class_name)



from typing import Optional

import inflect

_ENGINE: Optional[inflect.engine] = None


def get_inflect_engine():
    global _ENGINE
    if _ENGINE: return _ENGINE

    _ENGINE = inflect.engine()

    return _ENGINE



from crudhex.adapters.infrastructure.template_writer.adapters import domain_code_writer, db_code_writer, rest_code_writer
from crudhex.adapters.infrastructure.loader import fs_spec_loader as spec_loader


from pathlib import Path
from typing import Optional

import yaml


from crudhex.domain.models import ProjectConfig


def load_project_config(path: Path) -> ProjectConfig:
    config_data = _load_yaml_data(path)

    return ProjectConfig.from_dict(config_data)


def load_spec_config(path: Path) -> dict:
    return _load_yaml_data(path)


def _load_yaml_data(path: Path) -> dict:
    config_data: dict = {}
    with open(path.resolve(), 'r', encoding='utf-8') as f:
        config_data = yaml.safe_load(f)

    return config_data


def _default_path(file: str) -> Path:
    return Path('.') / file



from pathlib import Path
from typing import Dict, List, Optional, Union

from crudhex.domain.models import RelationType
from crudhex.adapters.infrastructure.template_writer.config import template_config
from crudhex.adapters.infrastructure.template_writer.config.template_config import get_db_file_path
from crudhex.adapters.infrastructure.template_writer.services.template_env import get_template_environment
from crudhex.adapters.infrastructure.template_writer.services.inflect_engine import get_inflect_engine


def create_entity(dest: Path, class_type: str, package: str,
                  imports: List[str], meta: Dict[str, str], fields: List[Dict[str, str]]):

    template_env = get_template_environment()

    id_field = _get_id_field(fields)
    if not id_field: raise RuntimeError('Id field is mandatory for db entity generation')

    fields_fragment = _generate_fields_fragment(fields)
    setters_fragment = _generate_fields_relation_setters(fields)

    entity_template = template_env.get_template(get_db_file_path(template_config.DB_ENTITY_TEMPLATE))
    entity_code = entity_template.render({
        'package': package,
        'imports': '\n'.join(imports),
        'class_type': class_type,
        'fields': fields_fragment,
        'methods': setters_fragment,
        'id_field': id_field['name'],
        **meta
    })

    with open(dest.resolve(), 'w+', encoding='utf-8') as f:
        f.write(entity_code)


def create_entity_repository(dest: Path, class_type: str, package: str,
                             imports: List[str], entity_type: str, id_type: str):

    template_env = get_template_environment()

    repository_template = template_env.get_template(get_db_file_path(template_config.DB_REPOSITORY_TEMPLATE))
    repository_code = repository_template.render({
        'package': package,
        'imports': '\n'.join(imports),
        'class_type': class_type,
        'entity_type': entity_type,
        'id_type': id_type
    })

    with open(dest.resolve(), 'w+', encoding='utf-8') as f:
        f.write(repository_code)


def create_adapter(dest: Path, class_type: str, package: str, class_type_interface: str,
                   imports: List[str], model_type: str, entity_type: str, id_type: str,
                   repository_type: str, create_command_type: str, update_command_type: str,
                   relations_data: List[Dict[str, str]], mapper_type: Optional[str] = None):

    template_env = get_template_environment()

    process_cmd_template = template_env.get_template(get_db_file_path(template_config.PROCESS_COMMAND))
    process_cmd_code = process_cmd_template.render({
        'entity_type': entity_type,
        'update_command_type': update_command_type,
        'command_relations': relations_data
    })

    adapter_template = template_env.get_template(get_db_file_path(template_config.DB_ADAPTER_TEMPLATE))
    adapter_code = adapter_template.render({
        'package': package,
        'imports': '\n'.join(imports),
        'class_type': class_type,
        'class_type_interface': class_type_interface,
        'main_repository_type': repository_type,
        'additional_repositories': _get_additional_repositories(relations_data, repository_type),
        'mapper_type': mapper_type,
        'model_type': model_type,
        'entity_type': entity_type,
        'id_type': id_type,
        'create_command_type': create_command_type,
        'update_command_type': update_command_type,
        'process_command': process_cmd_code
    })

    with open(dest.resolve(), 'w+', encoding='utf-8') as f:
        f.write(adapter_code)


def _generate_fields_fragment(fields: List[Dict[str, Union[str, RelationType]]]) -> str:
    template_env = get_template_environment()

    templates = {}
    field_fragments = []
    for field in fields:
        if field['relationship']:
            template_name = template_config.get_relation_template(field['relationship'], field['mapped_by'] is None)
            template = templates.setdefault(
                template_name,
                template_env.get_template(get_db_file_path(template_name))
            )

            field_fragments.append(template.render(field))
        else:
            template = templates.setdefault(
                template_config.FIELD,
                template_env.get_template(get_db_file_path(template_config.FIELD))
            )

            field_fragments.append(template.render(field))

    return '\n\n'.join(field_fragments)


def _generate_fields_relation_setters(fields: List[Dict[str, Union[RelationType, str]]]) -> str:
    template_env = get_template_environment()

    templates = {}
    setter_fragments = []
    for field in fields:
        if not field['relationship']: continue  # Sync setters only for relations

        template_name = template_config.get_sync_relation_template(field['relationship'], field['mapped_by'] is None)
        if not template_name: continue

        template = templates.setdefault(template_name, template_env.get_template(get_db_file_path(template_name)))
        params = {**field, 'field': field['name']}
        if field['relationship'].has_multiple():
            params['field_sing'] = get_inflect_engine().singular_noun(field['name'])

        setter_fragments.append(template.render(params))

    return '\n\n'.join(setter_fragments)


def _get_id_field(fields: List[Dict[str, str]]) -> Optional[Dict[str, str]]:
    id_field = None
    for field in fields:
        if field['id']:
            id_field = field
            break

    return id_field


def _get_additional_repositories(relations_data: List[Dict[str, str]], main_repository: str) -> List[str]:
    return [rd.get('repository_type') for rd in relations_data
            if rd.get('repository_type') and rd.get('repository_type') != main_repository]



from collections import namedtuple
from typing import List

from crudhex.domain.models import Field
from crudhex.domain.models.field import ClassType
from crudhex.domain.services.type_resolver import get_type_resolver
from crudhex.domain.utils.package_utils import generate_import, full_class_name


def get_field_imports(field: Field) -> List[str]:
    type_resolver = get_type_resolver()
    if field.type.is_native() or field.type.is_generated: return []

    class_types = type_resolver.get_field_types_full_class(field)

    return [generate_import(ct) for ct in class_types]


def get_type_import(class_type: ClassType) -> str:
    return generate_import(class_type.get_full_class_type())


def get_import(package: str, class_type: str) -> str:
    full_class = full_class_name(package, class_type)

    return generate_import(full_class)


def get_resolved_import(class_type: str) -> str:
    type_resolver = get_type_resolver()
    full_class = type_resolver.resolve_type(class_type)

    return generate_import(full_class)


def get_field_types(field: Field) -> 'FieldTypes':

    collection_type = None
    collection_type_impl = None
    if field.type.is_collection():
        type_resolver = get_type_resolver()
        collection_type = field.type.collection_type
        collection_type_impl = type_resolver.get_collection_type_impl(field.type.collection_type)

    return FieldTypes(field.type.class_type, collection_type, collection_type_impl)


FieldTypes = namedtuple('FieldTypes', 'class_type collection_type collection_type_impl')



from pathlib import Path
from typing import List, Dict, Optional

from crudhex.adapters.infrastructure.template_writer.services.template_env import get_template_environment
from crudhex.adapters.infrastructure.template_writer.config import template_config
from crudhex.adapters.infrastructure.template_writer.config.template_config import get_domain_file_path


def create_command(dest: Path, class_type: str, package: str,
                   imports: List[str], fields: List[Dict[str, str]]):

    _create_data_class(template_config.COMMAND_TEMPLATE, dest, class_type,
                       package, imports, fields)


def create_model(dest: Path, class_type: str, package: str,
                 imports: List[str], fields: List[Dict[str, str]]):

    _create_data_class(template_config.MODEL_TEMPLATE, dest, class_type,
                       package, imports, fields)


def create_db_port(dest: Path, class_type: str, package: str,
                   imports: List[str], id_type: str, model_type: str,
                   create_cmd_type: str, update_cmd_type: str):

    _create_port_class(template_config.DB_PORT_TEMPLATE, dest, class_type,
                       package, imports, id_type,
                       model_type, create_cmd_type, update_cmd_type)


def create_use_case_port(dest: Path, class_type: str, package: str,
                         imports: List[str], id_type: str, model_type: str,
                         create_cmd_type: str, update_cmd_type: str):

    _create_port_class(template_config.USE_CASE_PORT_TEMPLATE, dest, class_type,
                       package, imports, id_type,
                       model_type, create_cmd_type, update_cmd_type)


def create_use_case(dest: Path, class_type: str, package: str,
                    imports: List[str], id_type: str, model_type: str,
                    create_cmd_type: str, update_cmd_type: str, db_port_type: str):

    extras = {'db_port_type': db_port_type}

    _create_port_class(template_config.USE_CASE_TEMPLATE, dest, class_type,
                       package, imports, id_type,
                       model_type, create_cmd_type, update_cmd_type, extras)


def _create_data_class(template: str, dest: Path, class_type: str, package: str,
                       imports: List[str], fields: List[Dict[str, str]]):

    template_env = get_template_environment()

    fields_fragment = _generate_fields_fragment(fields)

    model_template = template_env.get_template(get_domain_file_path(template))
    model_code = model_template.render({
        'package': package,
        'imports': '\n'.join(imports),
        'class_type': class_type,
        'fields': fields_fragment
    })

    with open(dest.resolve(), 'w+', encoding='utf-8') as f:
        f.write(model_code)


def _create_port_class(template: str, dest: Path, class_type: str, package: str,
                       imports: List[str], id_type: str, model_type: str,
                       create_cmd_type: str, update_cmd_type: str, extra_params: Optional[Dict[str, str]] = None):

    template_env = get_template_environment()

    port_template = template_env.get_template(get_domain_file_path(template))
    port_data = {
        'package': package,
        'imports': '\n'.join(imports),
        'class_type': class_type,
        'model_type': model_type,
        'id_type': id_type,
        'create_command_type': create_cmd_type,
        'update_command_type': update_cmd_type
    }
    if extra_params: port_data.update(**extra_params)

    port_code = port_template.render(port_data)

    with open(dest.resolve(), 'w+', encoding='utf-8') as f:
        f.write(port_code)


def _generate_fields_fragment(fields: List[Dict[str, str]]):
    template_env = get_template_environment()
    field_template = template_env.get_template(get_domain_file_path(template_config.DOM_FIELD))

    field_fragments = []
    for field in fields:
        field_fragments.append(field_template.render(**field))

    return ''.join(field_fragments)



from pathlib import Path
from typing import List, Dict

from crudhex.domain.models import Entity, Field
from crudhex.domain.utils.class_type_utils import get_import, get_type_import, get_resolved_import
from . import entity_generator, repository_generator
from ..config_context import get_config
from ..domain import command_generator, model_generator, db_port_generator
from ...models.RelationData import RelationData
from ...models.field import ClassType
from crudhex.domain.ports import db_code_writer

_ADAPTER_PREFIX = 'DatasourceAdapter'
_DEFAULT_MAPPER_TYPE = 'Mapper'


def create_adapter_class(entity: Entity, entities_map: Dict[str, Entity], folder: Path) -> Path:
    if not folder.is_dir(): raise RuntimeError('Output path must be a folder ({})'.format(folder.resolve()))

    class_type = get_adapter_type_name(entity)
    adapter_file = folder / f'{class_type}.java'

    model_type = model_generator.get_model_type_name(entity)
    entity_type = entity_generator.get_entity_type_name(entity)
    id_type = entity.get_id_field().type.class_type
    command_type = command_generator.get_command_type_name(entity)
    repository_type = repository_generator.get_repository_type_name(entity)
    db_port = db_port_generator.get_port_type_name(entity)
    relations_data = [d.to_dict() for d in _get_relations_data(entity, entities_map)]

    db_code_writer.create_adapter(adapter_file, class_type, get_package(), db_port,
                                  _get_adapter_imports(entity, entities_map), model_type, entity_type, id_type,
                                  repository_type, command_type, command_type,
                                  relations_data, _get_mapper_type().class_type)

    return adapter_file


def get_package() -> str:
    return get_config().domain_use_cases_pkg


def get_adapter_type_name(entity: Entity) -> str:
    return f'{entity.name}{_ADAPTER_PREFIX}'


def _get_mapper_type() -> ClassType:
    config_mapper = get_config().db_mapper_class
    if not config_mapper:
        return ClassType(_DEFAULT_MAPPER_TYPE, True)

    return ClassType(config_mapper)


def _get_adapter_imports(entity: Entity, entities_map: Dict[str, Entity]) -> List[str]:
    imports = set(db_port_generator._get_port_imports(entity))
    imports.update([
        get_import(entity_generator.get_package(), entity_generator.get_entity_type_name(entity)),
        get_import(repository_generator.get_package(), repository_generator.get_repository_type_name(entity)),
        get_import(db_port_generator.get_package(), db_port_generator.get_port_type_name(entity))

    ])

    mapper_type = _get_mapper_type()
    if not mapper_type.is_generated:
        imports.add(get_type_import(mapper_type))

    # Getting updated relations classes (entity and repository)
    for field in _get_command_relation_fields(entity):
        related_entity = entities_map.get(field.type.class_type)
        if not related_entity: continue

        imports.update([
            get_import(entity_generator.get_package(), entity_generator.get_entity_type_name(related_entity)),
            get_import(repository_generator.get_package(), repository_generator.get_repository_type_name(related_entity))
        ])
        if field.type.collection_type:
            imports.add(get_resolved_import(field.type.collection_type))

    return list(imports)


def _get_relations_data(entity: Entity, entities_map: Dict[str, Entity]) -> List[RelationData]:
    relations_data = []
    for field in _get_command_relation_fields(entity):
        related_entity = entities_map.get(field.type.class_type, Entity(field.type.class_type))

        relation_data = RelationData(field.name)
        relation_data.collection_type = field.type.collection_type
        relation_data.entity_type = entity_generator.get_entity_type_name(related_entity)
        relation_data.repository_type = repository_generator.get_repository_type_name(related_entity)
        relation_data.create_command_type = command_generator.get_command_type_name(related_entity)
        relation_data.update_command_type = command_generator.get_command_type_name(related_entity)

        relations_data.append(relation_data)

    return relations_data


def _get_command_relation_fields(entity: Entity) -> List[Field]:
    return [f for f in command_generator.get_command_used_fields(entity) if f.has_relation()]



from pathlib import Path
from typing import List

from crudhex.domain.models import Entity
from crudhex.domain.utils.class_type_utils import get_field_imports, get_import
from ..config_context import get_config
from . import command_generator
from . import model_generator

from crudhex.domain.ports import domain_code_writer

_PORT_PREFIX = 'DatasourcePort'


def create_port_class(entity: Entity, folder: Path) -> Path:
    if not folder.is_dir(): raise RuntimeError('Output path must be a folder ({})'.format(folder.resolve()))

    class_type = get_port_type_name(entity)
    port_file = folder / f'{class_type}.java'

    id_type = entity.get_id_field().type.class_type
    model_type = model_generator.get_model_type_name(entity)
    command_type = command_generator.get_command_type_name(entity)

    domain_code_writer.create_db_port(port_file, class_type, get_package(),
                                      _get_port_imports(entity), id_type, model_type,
                                      command_type, command_type)

    return port_file


def get_package() -> str:
    return get_config().domain_out_ports_pkg


def get_port_type_name(entity: Entity) -> str:
    return f'{entity.name}{_PORT_PREFIX}'


def _get_port_imports(entity: Entity) -> List[str]:
    imports = [
        get_import(command_generator.get_package(), command_generator.get_command_type_name(entity)),
        get_import(model_generator.get_package(), model_generator.get_model_type_name(entity))
    ]
    imports += get_field_imports(entity.get_id_field())

    return imports



from pathlib import Path
from typing import Dict, Optional

from crudhex.domain.utils.package_utils import pkg_to_path


class ProjectConfig:
    src: Optional[str]
    _domain_src: Optional[str]
    _db_adapter_src: Optional[str]
    _rest_adapter_src: Optional[str]

    domain_models_pkg: str
    domain_commands_pkg: str
    domain_in_ports_pkg: str
    domain_out_ports_pkg: str

    db_models_pkg: str
    db_repositories_pkg: str
    db_adapters_pkg: str
    db_mapper_class: Optional[str]

    rest_models_pkg: str
    rest_controllers_pkg: str

    def __init__(self):
        self.src = None
        self._domain_src = None
        self._db_adapter_src = None
        self._rest_adapter_src = None

        self.domain_models_pkg = ''
        self.domain_commands_pkg = ''
        self.domain_in_ports_pkg = ''
        self.domain_out_ports_pkg = ''
        self.domain_use_cases_pkg = ''

        self.db_models_pkg = ''
        self.db_repositories_pkg = ''
        self.db_adapters_pkg = ''
        self.db_mapper_class = None

        self.rest_models_pkg = ''
        self.rest_controllers_pkg = ''

    @classmethod
    def from_dict(cls, data: Dict[str, str]) -> 'ProjectConfig':
        project_config = ProjectConfig()

        data = ProjectConfig._process_config(data)
        project_config.__dict__.update(**data)

        return project_config

    def get_domain_models_path(self) -> Path:
        return pkg_to_path(self.domain_models_pkg, self.domain_src)

    def get_domain_commands_path(self) -> Path:
        return pkg_to_path(self.domain_commands_pkg, self.domain_src)

    def get_domain_in_ports_path(self) -> Path:
        return pkg_to_path(self.domain_in_ports_pkg, self.domain_src)

    def get_domain_out_ports_path(self) -> Path:
        return pkg_to_path(self.domain_out_ports_pkg, self.domain_src)

    def get_domain_use_cases_path(self) -> Path:
        return pkg_to_path(self.domain_use_cases_pkg, self.domain_src)

    def get_db_models_path(self) -> Path:
        return pkg_to_path(self.db_models_pkg, self.db_adapter_src)

    def get_db_repositories_path(self) -> Path:
        return pkg_to_path(self.db_repositories_pkg, self.db_adapter_src)

    def get_db_adapters_path(self) -> Path:
        return pkg_to_path(self.db_adapters_pkg, self.db_adapter_src)

    def get_rest_models_path(self) -> Path:
        return pkg_to_path(self.rest_models_pkg, self.rest_adapter_src)

    def get_rest_controllers_path(self) -> Path:
        return pkg_to_path(self.rest_controllers_pkg, self.rest_adapter_src)

    @property
    def domain_src(self):
        if self._domain_src is None:
            return self.src

        return self._domain_src

    @domain_src.setter
    def domain_src(self, src: str):
        self._domain_src = src

    @property
    def db_adapter_src(self):
        if self._db_adapter_src is None:
            return self.src

        return self._db_adapter_src

    @db_adapter_src.setter
    def db_adapter_src(self, src: str):
        self._db_adapter_src = src

    @property
    def rest_adapter_src(self):
        if self._rest_adapter_src is None:
            return self.src

        return self._rest_adapter_src

    @rest_adapter_src.setter
    def rest_adapter_src(self, src: str):
        self._rest_adapter_src = src

    @classmethod
    def _process_config(cls, config_data: Dict[str, str]) -> Dict[str, str]:
        p_config = {}
        for key, val in config_data.items():
            p_config[key.replace('-', '_')] = val

        return p_config



from typing import Dict


class RelationData:
    name: str
    repository_type: str
    collection_type: str
    entity_type: str
    create_command_type: str
    update_command_type: str

    def __init__(self, name: str):
        self.name = name
        self.repository_type = None
        self.collection_type = None
        self.entity_type = None
        self.create_command_type = None
        self.update_command_type = None

    def is_collection(self) -> bool:
        return bool(self.collection_type)

    def to_dict(self) -> Dict[str, str]:
        data = self.__dict__
        data['is_collection'] = self.is_collection()

        return data



from enum import Enum
from typing import Optional
import uuid

from crudhex.domain.utils.package_utils import parse_class_name, full_class_name


class Field:
    name: str
    type: 'ClassType'
    column: Optional[str]
    id_meta: Optional['IdMeta']
    relation: Optional['Relation']

    def __init__(self, name: str, class_type: str, is_generated: bool = False):
        self.name = name
        self.type = ClassType(class_type, is_generated)
        self.column = None
        self.id_meta = None
        self.relation = None

    def is_id(self) -> bool:
        return self.id_meta is not None

    def has_relation(self) -> bool:
        return self.relation is not None

    def get_column_or_default(self) -> str:
        if not self.column: return self.name.lower()

        return self.column


class Relation:
    type: 'RelationType'
    join_table: Optional[str]
    join_column: Optional[str]
    inverse_join_column: Optional[str]
    mapped_by: Optional[str]

    def __init__(self, relation_type: 'RelationType'):
        self.type = relation_type
        self.join_table = None
        self.join_column = None
        self.inverse_join_column = None

    def to_dict(self):
        return self.__dict__


class IdMeta:
    generation: 'GenerationType'
    sequence: Optional[str]

    def __init__(self, generation: 'GenerationType'):
        self.generation = generation
        self.sequence = None

    def get_sequence_or_default(self, id_name: Optional[str] = None) -> str:
        if self.sequence: return self.sequence

        if not id_name:
            id_name = uuid.uuid4()

        return f'{id_name}_pk_gen'

    def to_dict(self):
        data = self.__dict__
        data['sequence'] = self.get_sequence_or_default()

        return data


class ClassType:
    package: Optional[str]
    class_type: str
    collection_type: Optional[str]
    is_generated: bool

    def __init__(self, class_name: str, is_generated: bool = False):
        class_data = parse_class_name(class_name)
        self.package = class_data[0]
        self.class_type = class_data[1]
        self.collection_type = None
        self.is_generated = is_generated

    def get_full_class_type(self) -> str:
        return full_class_name(self.package, self.class_type)

    def is_native(self) -> bool:
        return self.class_type[0].islower()

    def is_collection(self) -> bool:
        return self.collection_type is not None


class GenerationType(Enum):
    NONE = 'none'
    SEQUENCE = 'sequence'
    IDENTITY = 'identity'
    AUTO = 'auto'


class RelationType(Enum):
    ONE_TO_ONE = 'one-to-one'
    ONE_TO_MANY = 'one-to-many'
    MANY_TO_ONE = 'many-to-one'
    MANY_TO_MANY = 'many-to-many'

    def has_multiple(self) -> bool:
        return self in [RelationType.ONE_TO_MANY, RelationType.MANY_TO_MANY]



from typing import Optional

from jinja2 import Environment, PackageLoader

from ..config import template_config

_TEMPLATE_ENV: Optional[Environment] = None


def get_template_environment():
    global _TEMPLATE_ENV
    if _TEMPLATE_ENV: return _TEMPLATE_ENV

    _TEMPLATE_ENV = Environment(loader=PackageLoader(template_config.TEMPLATES_PACKAGE, template_config.TEMPLATE_FOLDER),
                                trim_blocks=True, lstrip_blocks=True)

    _TEMPLATE_ENV.filters['firstlower'] = _first_lower_filter
    _TEMPLATE_ENV.filters['firstupper'] = _first_upper_filter

    return _TEMPLATE_ENV


def _first_lower_filter(value: str) -> str:
    if not value or len(value) == 0: return value

    return value[0].lower() + value[1:]


def _first_upper_filter(value: str) -> str:
    if not value or len(value) == 0: return value

    return value[0].upper() + value[1:]
